
; MzPlasm
; Author: Simone Portuesi
; DATE: January 2007
;==================================================

; types

; usefull subype predicate
(require (lib "43.ss" "srfi"))

; Curried version of vector-every but returns false if not a vector
(define (vector-every2 p) (lambda (v) (and (vector? v) (vector-every p v))))

; Curried version of vector-any but returns false if not a vector
(define (vector-any2 p) (lambda (v) (and (vector? v) (vector-any p v))))

; On a list of vectors and values, 
; returns the i-th element if vector, the value otherwise
(define (vectors-ref vectors i)
  (map (lambda (v) (if (vector? v) (vector-ref v i) v)) vectors))

; On a list, checks if p applies to one or more element
; if yes it applies checks if f returns same value, 
;                          if yes returns the value
;                          if no returns 'neq
; if not returns 'none
(define (any-f p f l)
  (do ((todo l (cdr todo)) 
       (res 'none (if (p (car todo))
		      (cond ((eq? res 'none)  (f (car todo)))
			    ((eq? res 'neq)   'neq)
			    (#t (if (eq? res (f (car todo)))
				    res
				    'neq)))
		      res)))
      ((or (null? todo) (eq? res 'neq)) res)))

; same as any-f
; but in case p fails op is checked
(define (any-p-f p op f l)
  (do ((todo l (cdr todo)) 
       (res 'none (cond ((p (car todo))
			 (cond ((eq? res 'none)  (f (car todo)))
			       ((eq? res 'neq)   'neq)
			       (#t (if (eq? res (f (car todo)))
				       res
				       'neq))))
			((op (car todo)) res)
			(#f 'other))))
      ((or (null? todo) (eq? res 'neq) (eq? res 'other)) res)))


; If arg is a vector, it maps f on each element 
; applies f to arg otherwise
(define (KL-f1 f)
  (lambda (arg)
    (if (vector? arg)
	(let* ((mf (lambda (i e) (f e)))
	       (res (vector-map mf arg))
	       (first (vector-ref res 0)))
	       (if (vector-every (lambda (e) (eqv? first e)) res)
		   first
		   res))
	(f arg))))

; Same as KL-f1, but checks if every element of the vector satisfies p
; if not (e arg) is returned
(define (KL-p-f1 p f e)
  (lambda (arg)
    (if (vector? arg) 
	(if ((vector-every2 p) arg)
	    (let* ((mf (lambda (i e) (f e)))
		   (res (vector-map mf arg))
		   (first (vector-ref res 0)))
	      (if (vector-every (lambda (e) (eqv? first e)) res)
		  first
		  res))
	    (e arg)) ;sarebbe da individuare il/i responsabili
	(if (p arg)
	    (f arg)
	    (e arg)))))

; If args contains a a vector, it returns a vector
; On a list of vectors and values, 
; returns (apply f args) if all values
; otherwise it returns a vector containig all applications to
; the i-th element if vector, the value otherwise
(define (KL-fn f . e)
  (lambda (args)
    (if (pair? args)
	(let*  ((vlen (any-f vector?  vector-length args)))
	  (cond ((number? vlen) 
		 (let* ((res  
			 (vector-unfold (lambda (i) (f (vectors-ref args i)))
					vlen))
			(first (vector-ref res 0)))
		   (if (vector-every (lambda (e) (eqv? first e)) res)
		       first
		       res)))
		((eq? vlen 'none) (f args))
		(#t  (error 'KL-fn "Non homogeneus" args))))
	(if (null? e) args ((car e) args)))))

(define (KL-p-fn p f . e)
  (lambda (args)
    (if (pair? args)
	(let*  ((vlen (any-p-f vector? p vector-length args)))
	  (cond ((number? vlen) 
		 (let* ((res  
			 (vector-unfold (lambda (i) (f (vectors-ref args i)))
					vlen))
			(first (vector-ref res 0)))
		   (if (vector-every (lambda (e) (eqv? first e)) res)
		       first
		       res)))
		((eq? vlen 'none) (f args))
		((eq? vlen 'other) (e args))
		(#t  (error 'KL-fn "KL-p-fn" args))))
	(if (null? e) args ((car e) args)))))

; same as KL-fn but checks a global predicate first
; note that the predicate should just check the factored structure
; for predicates on each element use next function
(define (KL-tp-fn tp f e)
  (lambda (args)
    (if  (tp args)
	 ((KL-fn f) args)
	 (e args))))

; same as KL-fn, but checks p on each element of the list
; adapting in case is a vector or a single value
; not quite working p should be cheched on values as well ...
;; (define (KL-p-fn p f e)
;;   (lambda (args)
;;     (if (or (not p) (every p args)) 
;; 	(apply f args)
;; 	(let*  ((vlen (if p 
;; 			  (any-f (vector-every2 p) vector-length args)
;; 			  (any-f vector?  vector-length args))))
;; 	  (if (number? vlen) 
;; 	      (let* ((res  
;; 		      (vector-unfold (lambda (i) (apply f (vectors-ref args i)))
;; 				     vlen))
;; 		     (first (vector-ref res 0)))
;; 		(if (vector-every (lambda (e) (eqv? first e)) res)
;; 		    first
;; 		    res))
;; 	      (e args))))))



(define (KL-apl-fn p f e)
  (lambda (args)
    (if (pair? args)
	(if (every pair? args)
	    (if (every (ceq (length (car args))) (map length (cdr args)))
                ; quanto sei brutta
		(apply map `(,(lambda arg ((KL-apl-fn p f e) arg)) ,@args)) 
		(e args))
	    ((KL-p-fn p f) args))
	args)))
  
(define (KL->ML L)
  (let ((res
 	 (cond ((pair? L) 
		(let* ((recstep (map KL->ML L))
		       (switch (any-f vector? vector-length recstep)))
		  
		  
		  (cond ((eq? switch 'none) recstep)
			((integer? switch) (vector-unfold 
					     (lambda (i) (vectors-ref recstep i) )
					     switch))
			 (#t (error 'KL->ML "KL->ML" L))
			 )))
	       ((vector? L) L)
	       (#t  L)) 
 	 ))
    (if (and (memq 'fmap *DEBUG*)  (memq 'KL->ML *DEBUG*)) 
	(Plasm_Printer `(KL->ML ,L "->" ,res) #f))
    res
    ))


(define (ML->KL ML)
  (let ((res
	 (if (vector? ML)
	     
	     (let* ((len (vector-length ML))
		    (first (vector-ref ML 0))
		    (clen (if (list? first) (length first) -1)))
	       
	       ;(Plasm_Printer `(ML-KL "in" ,len ,first ,clen) #f)
	       (cond ((eq? clen 0) '())
		     ((eq? clen -1) (if (vector-every (lambda (e) (eqv? first e)) ML)
					first
					ML))
		     (#t (if (vector-every (lambda (e) (eq? clen (length e))) ML)
			     (do ((i   1   (+ i 1))
				  (res '() (cons (ML->KL (vector-map (lambda (j e) (car e)) workV)) res))
				  (workV ML (vector-map! (lambda (j e) (cdr e)) workV)))
				 ((> i clen) (reverse res))
			       ;(Plasm_Printer `("step" ,i ,res ,workV) #f))
			       )
			     ML))))
	      
	     (error 'ML->KL "ML->KL" ML)))) ;should not be the case 
	 
  
     ; (if (and (memq 'fmap *DEBUG*)  (memq 'ML->KL *DEBUG*)) 
	 ; (Plasm_Printer `(ML->KL ,ML "->" ,res) #f))
      res)
  )
  



(define (fmap-fun-MAP fun)
  (lambda (vals)
    (if (memq 'fmap *DEBUG*) 
	(Plasm_Printer `(fmap ,fun ,vals) #t))
    
    (let* ((res (time (fun vals)))
	   (d   (if (memq 'fmap *DEBUG*) 
		    (Plasm_Printer `(fmap fun ,res1) #t))) )
      
      (if  (every (lambda (e) (or (vector? e) (number? e))) res)
	   res
	     (plasm-exception `( "FMAP" "Result " ,res1)))
      )
    )
  )
  
    


(define (fmap-fun-AA fun vals)

   (if (memq 'fmap *DEBUG*) 
       (Plasm_Printer `(fmap ,fun ,vals) #t))
 
   (let* ((res1 (list->vector vals))
	  (len (vector-length res1))
	  (d1   (if (memq 'fmap *DEBUG*) 
		    (Plasm_Printer `(fmap make-ML ,res1) #t)))
	  (res2 (time (ML->KL res1)))
	  (d2   (if (memq 'fmap *DEBUG*) 
		    (Plasm_Printer `(fmap ML-KL ,res2) #t)))
	  (res3  (time (fun res2)))
	  (d3   (if (memq 'fmap *DEBUG*) 
		    (Plasm_Printer `(fmap fun ,res3) #t)))
	  (res4 (time (KL->ML res3)))
	  (d4   (if (memq 'fmap *DEBUG*) 
		    (Plasm_Printer `(fmap KL->ML ,res4) #t)))
	  )
      
     (if (vector? res4)
	 (vector->list res4)
	 (do ((i 0 (+ i 1))
	      (res '() (cons res4 res)))
	     ((>= i len) res)))
     
     ))

